"""Registry object common."""
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, Generic, Optional, TypeVar, Union, cast

from asyncstdlib import cache
from fred_api.registry.contact.contact_info_types_pb2 import ContactIdRequest
from fred_api.registry.domain.domain_info_types_pb2 import DomainIdRequest
from fred_api.registry.keyset.keyset_info_types_pb2 import KeysetIdRequest
from fred_api.registry.nsset.nsset_history_types_pb2 import NssetHistoryRequest
from fred_api.registry.nsset.nsset_info_types_pb2 import NssetIdRequest, NssetInfoRequest
from fred_api.registry.nsset.nsset_state_history_types_pb2 import NssetStateHistoryRequest
from fred_api.registry.nsset.nsset_state_types_pb2 import NssetStateRequest
from fred_types import BaseModel, BaseObjectHistoryId, BaseObjectId
from frgal.aio import AsyncGrpcClient
from google.protobuf.empty_pb2 import Empty

from regal.utils import AbcAttribute

ObjectIdT = TypeVar("ObjectIdT", bound=BaseObjectId)
ObjectHistoryIdT = TypeVar("ObjectHistoryIdT", bound=BaseObjectHistoryId)


class ObjectStateFlag(BaseModel):
    """Represents properties of an object state flag.

    Attributes:
        manual: Whether the state flag is set manually.
        internal: Whether the state flag is internal.
    """

    manual: bool
    internal: bool


class RegistryObjectClient(ABC, AsyncGrpcClient, Generic[ObjectIdT, ObjectHistoryIdT]):
    """Abstract service gRPC client for registry objects."""

    service: str = AbcAttribute
    object_id: str = AbcAttribute
    object_history_id: str = AbcAttribute
    object_handle: str = AbcAttribute

    async def get_object_id(
        self,
        request: Union[ContactIdRequest, DomainIdRequest, KeysetIdRequest, NssetIdRequest],
        method: str,
        handle: str,
    ) -> str:
        """Get registry object id."""
        getattr(request, self.object_handle).value = handle
        result = await self.call(self.service, method, request)
        return cast(str, result[self.object_id])

    async def get_object_info(
        self,
        request: Union[NssetInfoRequest],
        method: str,
        object_id: ObjectIdT,
        history_id: Optional[ObjectHistoryIdT] = None,
    ) -> Dict[str, Any]:
        """Get registry object info."""
        getattr(request, self.object_id).uuid.value = object_id
        if history_id is not None:
            getattr(request, self.object_history_id).uuid.value = history_id
        result = await self.call(self.service, method, request)
        return cast(Dict[str, Any], result)

    async def get_object_state(
        self,
        request: Union[NssetStateRequest],
        method: str,
        object_id: ObjectIdT,
        *,
        internal: bool = True,
    ) -> Dict[str, bool]:
        """Get object history state.

        Arguments:
            request: Basic gRPC request message.
            method: Name of the method to be called.
            object_id: Object identifier.
            internal: Whether to include internal states in the result.
        """
        getattr(request, self.object_id).uuid.value = object_id
        result = await self.call(self.service, method, request)
        state = cast(Dict[str, bool], result["state"])
        if not internal:
            flags = await self.get_state_flags()
            state = {k: v for k, v in state.items() if not flags[k].internal}
        return state

    async def get_object_history(
        self,
        request: Union[NssetHistoryRequest],
        method: str,
        object_id: ObjectIdT,
        start: Optional[Union[ObjectHistoryIdT, datetime]] = None,
        end: Optional[Union[ObjectHistoryIdT, datetime]] = None,
    ) -> Dict[str, Any]:
        """Get registry object history."""
        getattr(request, self.object_id).uuid.value = object_id
        self._set_history_limits(request, start, end)

        result = await self.call(self.service, method, request)
        return cast(Dict[str, Any], result)

    async def get_object_state_history(
        self,
        request: Union[NssetStateHistoryRequest],
        method: str,
        object_id: ObjectIdT,
        start: Optional[Union[ObjectHistoryIdT, datetime]] = None,
        end: Optional[Union[ObjectHistoryIdT, datetime]] = None,
    ) -> Dict[str, Any]:
        """Get registry object state history."""
        getattr(request, self.object_id).uuid.value = object_id
        self._set_history_limits(request, start, end)

        result = await self.call(self.service, method, request)
        return cast(Dict[str, Any], {self.object_id: result[self.object_id], **result["history"]})

    def _set_history_limits(
        self,
        request: Union[NssetHistoryRequest, NssetStateHistoryRequest],
        start: Optional[Union[ObjectHistoryIdT, datetime]] = None,
        end: Optional[Union[ObjectHistoryIdT, datetime]] = None,
    ) -> None:
        """Set history limits to request.

        Args:
            request: Request with `history` item of ContactHistoryInterval type.
            start: Datetime or history_id of history interval start.
            end: Datetime or history_id of history interval end.
        """
        if start is not None:
            if isinstance(start, datetime):
                request.history.lower_limit.timestamp.FromDatetime(start)
            else:
                getattr(request.history.lower_limit, self.object_history_id).uuid.value = start
        if end is not None:
            if isinstance(end, datetime):
                request.history.upper_limit.timestamp.FromDatetime(end)
            else:
                getattr(request.history.upper_limit, self.object_history_id).uuid.value = end

    @cache
    async def get_object_state_flags(
        self,
        method: str,
    ) -> Dict[str, ObjectStateFlag]:
        """Get object state flags and their properties."""
        return cast(Dict[str, ObjectStateFlag], await self.call(self.service, method, Empty()))

    @abstractmethod
    async def get_state_flags(self) -> Dict[str, ObjectStateFlag]:
        """Get object state flags and their properties."""
