from datetime import datetime
from typing import Any, Dict, List, Sequence
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import call, sentinel

import pytz
from aioitertools import iter as aiter, list as alist
from fred_api.registry.domain.domain_history_types_pb2 import (
    DomainHistoryRecord,
    DomainHistoryReply,
    DomainHistoryRequest,
)
from fred_api.registry.domain.domain_info_types_pb2 import (
    DomainIdReply,
    DomainIdRequest,
    DomainInfoReply,
    DomainInfoRequest,
)
from fred_api.registry.domain.domain_life_cycle_stage_types_pb2 import (
    DomainLifeCycleStageReply,
    DomainLifeCycleStageRequest,
)
from fred_api.registry.domain.domain_state_flags_types_pb2 import DomainStateFlagsReply
from fred_api.registry.domain.domain_state_history_types_pb2 import (
    DomainStateHistory,
    DomainStateHistoryReply,
    DomainStateHistoryRequest,
)
from fred_api.registry.domain.domain_state_types_pb2 import DomainStateReply, DomainStateRequest
from fred_api.registry.domain.domains_by_contact_types_pb2 import DomainsByContactReply, DomainsByContactRequest
from fred_api.registry.domain.service_admin_grpc_pb2 import (
    BatchDeleteDomainsReply,
    BatchDeleteDomainsRequest,
    DomainContactInfo,
    GetDomainsDeleteNotifyInfoRequest,
    GetDomainsNotifyInfoReply,
    GetDomainsOutzoneNotifyInfoRequest,
    ManageDomainStateFlagsReply,
    ManageDomainStateFlagsRequest,
    UpdateDomainsAdditionalNotifyInfoReply,
    UpdateDomainsDeleteAdditionalNotifyInfoRequest,
    UpdateDomainsOutzoneAdditionalNotifyInfoRequest,
)
from fred_types import (
    ContactId,
    ContactRef,
    DomainHistoryId,
    DomainId,
    DomainRef,
    KeysetId,
    KeysetRef,
    NssetId,
    NssetRef,
)
from frgal.utils import AsyncTestClientMixin, make_awaitable
from google.protobuf.empty_pb2 import Empty

from regal.common import DatetimeRange
from regal.domain import (
    ContactInfo,
    ContactInfoType,
    Domain,
    DomainAdminClient,
    DomainAdminDecoder,
    DomainClient,
    DomainContactRole,
    DomainDecoder,
    DomainLifeCycleStage,
    DomainLifeCycleStageResult,
    NotDeletedReason,
)
from regal.exceptions import DomainDoesNotExist
from regal.tests.common import ObjectClientTestMixin


class TestDomainClient(AsyncTestClientMixin, DomainClient):
    """Test DomainClient."""


class TestDomainAdminClient(AsyncTestClientMixin, DomainAdminClient):
    """Test DomainAdminClient."""

    exhaust_iterables = True


class DomainDecoderTest(TestCase):
    def setUp(self):
        self.decoder = DomainDecoder()

    def test_decode_domain_by_contact_no_role(self):
        domain = DomainsByContactReply.Data.Domain()
        domain.domain.id.uuid.value = "domain.1"
        domain.domain.fqdn.value = "example.com"
        domain.domain_history_id.uuid.value = "domain.history.1"
        domain.is_deleted = True
        self.assertEqual(
            self.decoder.decode(domain),
            {
                "domain": DomainRef(id=DomainId("domain.1"), fqdn="example.com"),
                "domain_history_id": DomainHistoryId("domain.history.1"),
                "roles": set(),
                "is_deleted": True,
            },
        )

    def test_decode_domain_by_contact_one_role(self):
        data = {
            "is_holder": DomainContactRole.HOLDER,
            "is_admin_contact": DomainContactRole.ADMIN,
            "is_keyset_tech_contact": DomainContactRole.KEYSET_TECH,
            "is_nsset_tech_contact": DomainContactRole.NSSET_TECH,
        }

        for name, role in data.items():
            with self.subTest(name=name, role=role):
                domain = DomainsByContactReply.Data.Domain()
                domain.domain.id.uuid.value = "domain.1"
                domain.domain.fqdn.value = "example.com"
                domain.domain_history_id.uuid.value = "domain.history.1"
                setattr(domain, name, True)
                self.assertEqual(
                    self.decoder.decode(domain),
                    {
                        "domain": DomainRef(id=DomainId("domain.1"), fqdn="example.com"),
                        "domain_history_id": DomainHistoryId("domain.history.1"),
                        "roles": {role},
                        "is_deleted": False,
                    },
                )

    def test_decode_domain_by_contact_all_roles(self):
        domain = DomainsByContactReply.Data.Domain()
        domain.domain.id.uuid.value = "domain.1"
        domain.domain.fqdn.value = "example.com"
        domain.domain_history_id.uuid.value = "domain.history.1"
        domain.is_holder = True
        domain.is_admin_contact = True
        domain.is_keyset_tech_contact = True
        domain.is_nsset_tech_contact = True
        self.assertEqual(
            self.decoder.decode(domain),
            {
                "domain": DomainRef(id=DomainId("domain.1"), fqdn="example.com"),
                "domain_history_id": DomainHistoryId("domain.history.1"),
                "roles": set(DomainContactRole),
                "is_deleted": False,
            },
        )


class DomainAdminDecoderTest(TestCase):
    def setUp(self):
        self.decoder = DomainAdminDecoder()

    def test_decode_domain_contact_info(self):
        self.assertEqual(
            self.decoder.decode(DomainContactInfo(email="rimmer@example.org")),
            ContactInfo(type=ContactInfoType.EMAIL, value="rimmer@example.org"),
        )
        self.assertEqual(
            self.decoder.decode(DomainContactInfo(telephone="123456")),
            ContactInfo(type=ContactInfoType.TELEPHONE, value="123456"),
        )
        self.assertIsNone(self.decoder.decode(DomainContactInfo()))


class DomainClientTest(ObjectClientTestMixin[TestDomainClient], IsolatedAsyncioTestCase):
    """Test DomainClient's methods."""

    grpc_namespace = "/Fred.Registry.Api.Domain.Domain"
    get_state_request = DomainStateRequest
    get_state_reply = DomainStateReply
    get_state_method = "get_domain_state"
    get_history_request = DomainHistoryRequest
    get_history_reply = DomainHistoryReply
    get_history_method = "get_domain_history"
    get_history_record = DomainHistoryRecord
    get_state_history_request = DomainStateHistoryRequest
    get_state_history_reply = DomainStateHistoryReply
    get_state_history_method = "get_domain_state_history"
    get_state_history_record = DomainStateHistory.Record
    get_state_flags_method = "get_domain_state_flags"
    get_state_flags_reply_cls = DomainStateFlagsReply

    def setUp(self):
        self.client = TestDomainClient(netloc=sentinel.netloc)

    async def _test_get_domain_id(self, fqdn: str, request_fqdn: str) -> None:
        domain_id = DomainId("kryten")
        reply = DomainIdReply()
        reply.data.domain_id.uuid.value = str(domain_id)
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_domain_id(fqdn)
        self.assertEqual(response, domain_id)
        self.assertIsInstance(response, DomainId)

        request = DomainIdRequest()
        request.fqdn.value = request_fqdn
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Domain.Domain/get_domain_id",
            timeout=None,
        )

    async def test_get_domain_id(self):
        await self._test_get_domain_id("example.org", "example.org")

    async def test_get_domain_id_idn(self):
        await self._test_get_domain_id("háčkyčárky.cz", "xn--hkyrky-ptac70bc.cz")

    async def _test_get_domain_info(self, domain_history_id: bool = False) -> None:
        reply = DomainInfoReply()
        reply.data.fqdn.value = "example.com"
        reply.data.domain_id.uuid.value = "holly"
        reply.data.domain_history_id.uuid.value = "hilly"

        reply.data.registrant.id.uuid.value = "JMC"
        reply.data.registrant.handle.value = "JMC-HANDLE"
        reply.data.registrant.name = "Arnold Rimmer"

        reply.data.nsset.id.uuid.value = "red-dwarf"
        reply.data.nsset.handle.value = "RED-DWARF"

        reply.data.keyset.id.uuid.value = "queeg-500"
        reply.data.keyset.handle.value = "QUEEG-500"

        reply.data.sponsoring_registrar.value = "REG-JMC"
        reply.data.expires_at.FromDatetime(datetime(2030, 1, 1, 10, 0, 4, tzinfo=pytz.utc))
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_domain_info(
            DomainId("holly"), history_id=DomainHistoryId("hilly") if domain_history_id else None
        )

        self.assertEqual(
            response,
            Domain(
                fqdn="example.com",
                domain_id=DomainId("holly"),
                domain_history_id=DomainHistoryId("hilly"),
                registrant=ContactRef(id=ContactId("JMC"), handle="JMC-HANDLE", name="Arnold Rimmer"),
                nsset=NssetRef(id=NssetId("red-dwarf"), handle="RED-DWARF"),
                keyset=KeysetRef(id=KeysetId("queeg-500"), handle="QUEEG-500"),
                sponsoring_registrar="REG-JMC",
                expires_at=datetime(2030, 1, 1, 10, 0, 4, tzinfo=pytz.utc),
            ),
        )

        request = DomainInfoRequest()
        request.domain_id.uuid.value = "holly"
        if domain_history_id:
            request.domain_history_id.uuid.value = "hilly"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Domain.Domain/get_domain_info",
            timeout=None,
        )

    async def test_get_domain_info(self):
        await self._test_get_domain_info(False)

    async def test_get_domain_info_with_history_id(self):
        await self._test_get_domain_info(True)

    async def test_get_domain_info_no_expiry_date(self):
        reply = DomainInfoReply()
        reply.data.fqdn.value = "example.com"
        reply.data.domain_id.uuid.value = "holly"
        reply.data.domain_history_id.uuid.value = "hilly"
        reply.data.registrant.id.uuid.value = "JMC"
        reply.data.registrant.handle.value = "JMC-HANDLE"
        reply.data.sponsoring_registrar.value = "REG-JMC"
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_domain_info(domain_id=DomainId("holly"))

        self.assertEqual(
            response,
            Domain(
                fqdn="example.com",
                domain_id=DomainId("holly"),
                domain_history_id=DomainHistoryId("hilly"),
                registrant=ContactRef(id=ContactId("JMC"), handle="JMC-HANDLE"),
                sponsoring_registrar="REG-JMC",
                expires_at=None,
            ),
        )

    async def test_get_domain_info_lifecycle(self):
        events = {
            "expiration_warning_scheduled_at",
            "outzone_unguarded_warning_scheduled_at",
            "outzone_scheduled_at",
            "delete_warning_scheduled_at",
            "delete_candidate_scheduled_at",
            "validation_expires_at",
            "outzone_at",
            "delete_candidate_at",
        }
        for event in events:
            with self.subTest(event=event):
                timestamp = datetime(1988, 9, 6, 12, tzinfo=pytz.UTC)
                reply = DomainInfoReply()
                reply.data.fqdn.value = "example.com"
                reply.data.domain_id.uuid.value = "holly"
                getattr(reply.data, event).FromDatetime(timestamp)
                self.client.mock.return_value = make_awaitable(reply)

                response = await self.client.get_domain_info(DomainId("holly"))

                domain = Domain(fqdn="example.com", domain_id=DomainId("holly"))
                setattr(domain, event, timestamp)
                self.assertEqual(response, domain)

    async def test_get_domains_by_contact_empty(self):
        reply = DomainsByContactReply()
        reply.data.domains.extend([])
        self.client.mock.return_value = aiter([reply])

        response = await self.client.get_domains_by_contact(contact_id=ContactId("contact.1"), roles=set())

        self.assertEqual(response, {"domains": []})
        request = DomainsByContactRequest()
        request.contact_id.uuid.value = "contact.1"
        self.assertEqual(
            self.client.mock.mock_calls,
            [
                call(request, method="/Fred.Registry.Api.Domain.Domain/get_domains_by_contact", timeout=None),
            ],
        )

    async def test_get_domains_by_contact_empty_with_pagination(self):
        reply = DomainsByContactReply()
        reply.data.domains.extend([])
        reply.data.pagination.items_left = 0
        self.client.mock.return_value = aiter([reply])

        response = await self.client.get_domains_by_contact(
            contact_id=ContactId("contact.1"),
            roles=set(),
            page_size=10,
        )

        self.assertEqual(response, {"domains": [], "pagination": {"next_page_token": "", "items_left": 0}})
        request = DomainsByContactRequest()
        request.contact_id.uuid.value = "contact.1"
        request.pagination.page_size = 10
        self.assertEqual(
            self.client.mock.mock_calls,
            [
                call(request, method="/Fred.Registry.Api.Domain.Domain/get_domains_by_contact", timeout=None),
            ],
        )

    async def test_get_domains_by_contact_full(self):
        reply = DomainsByContactReply()
        domain = reply.data.domains.add()
        domain.domain.id.uuid.value = "domain.1"
        domain.domain.fqdn.value = "domain.example.com"
        domain.domain_history_id.uuid.value = "domain.history.1"
        domain.is_deleted = True
        domain.is_holder = True
        domain.is_admin_contact = True
        reply.data.pagination.items_left = 1
        reply.data.pagination.next_page_token = "token.2"
        self.client.mock.return_value = aiter([reply])

        response = await self.client.get_domains_by_contact(
            contact_id=ContactId("contact.1"),
            roles=set(DomainContactRole),
            include_deleted=True,
            order_by="-fqdn",
            fqdn_filter="*dom*",
            page_token="token.1",
            page_size=1,
        )

        self.assertEqual(
            response,
            {
                "domains": [
                    {
                        "domain": DomainRef(id=DomainId("domain.1"), fqdn="domain.example.com"),
                        "domain_history_id": DomainHistoryId("domain.history.1"),
                        "roles": {DomainContactRole.HOLDER, DomainContactRole.ADMIN},
                        "is_deleted": True,
                    }
                ],
                "pagination": {"next_page_token": "token.2", "items_left": 1},
            },
        )
        request = DomainsByContactRequest()
        request.contact_id.uuid.value = "contact.1"
        request.include_deleted_domains = True
        request.order_by = "-fqdn"
        request.fqdn_filter = "*dom*"
        request.include_holder = True
        request.include_admin_contact = True
        request.include_keyset_tech_contact = True
        request.include_nsset_tech_contact = True
        request.pagination.page_token = "token.1"
        request.pagination.page_size = 1
        self.assertEqual(
            self.client.mock.mock_calls,
            [
                call(request, method="/Fred.Registry.Api.Domain.Domain/get_domains_by_contact", timeout=None),
            ],
        )

    async def test_get_domains_by_contact_chunks(self):
        reply_1 = DomainsByContactReply()
        domain_1 = reply_1.data.domains.add()
        domain_1.domain.id.uuid.value = "domain.1"
        domain_1.domain.fqdn.value = "domain.example.com"
        domain_1.domain_history_id.uuid.value = "domain.history.1"
        domain_1.is_deleted = True
        domain_1.is_holder = True
        domain_1.is_admin_contact = True
        reply_1.data.pagination.items_left = 1
        reply_1.data.pagination.next_page_token = "token.2"
        reply_2 = DomainsByContactReply()
        domain_2 = reply_2.data.domains.add()
        domain_2.domain.id.uuid.value = "domain.2"
        domain_2.domain.fqdn.value = "domain.example.org"
        domain_2.domain_history_id.uuid.value = "domain.history.2"
        domain_2.is_deleted = False
        domain_2.is_holder = False
        domain_2.is_admin_contact = True
        self.client.mock.return_value = aiter([reply_1, reply_2])

        response = await self.client.get_domains_by_contact(
            contact_id=ContactId("contact.1"),
            roles=set(DomainContactRole),
            include_deleted=True,
            order_by="-fqdn",
            fqdn_filter="*dom*",
            page_token="token.1",
            page_size=1,
        )

        self.assertEqual(
            response,
            {
                "domains": [
                    {
                        "domain": DomainRef(id=DomainId("domain.1"), fqdn="domain.example.com"),
                        "domain_history_id": DomainHistoryId("domain.history.1"),
                        "roles": {DomainContactRole.HOLDER, DomainContactRole.ADMIN},
                        "is_deleted": True,
                    },
                    {
                        "domain": DomainRef(id=DomainId("domain.2"), fqdn="domain.example.org"),
                        "domain_history_id": DomainHistoryId("domain.history.2"),
                        "roles": {DomainContactRole.ADMIN},
                        "is_deleted": False,
                    },
                ],
                "pagination": {"next_page_token": "token.2", "items_left": 1},
            },
        )

    async def _test_get_domain_life_cycle_stage(self, fqdn: str, request_fqdn: str) -> None:
        reply = DomainLifeCycleStageReply()
        reply.data.domain_id.uuid.value = "42"
        self.client.mock.return_value = make_awaitable(reply)

        result = await self.client.get_domain_life_cycle_stage(fqdn)

        self.assertEqual(result, "42")
        request = DomainLifeCycleStageRequest()
        request.fqdn.value = request_fqdn
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.Registry.Api.Domain.Domain/get_domain_life_cycle_stage", timeout=None)],
        )

    async def test_get_domain_life_cycle_stage(self):
        await self._test_get_domain_life_cycle_stage("example.org", "example.org")

    async def test_get_domain_life_cycle_stage_idn(self):
        await self._test_get_domain_life_cycle_stage("háčkyčárky.cz", "xn--hkyrky-ptac70bc.cz")

    async def test_get_domain_life_cycle_stage_delete_candidate(self):
        reply = DomainLifeCycleStageReply()
        reply.data.delete_candidate.CopyFrom(Empty())
        self.client.mock.return_value = make_awaitable(reply)

        result = await self.client.get_domain_life_cycle_stage("example.org")

        self.assertIsNone(result)
        request = DomainLifeCycleStageRequest()
        request.fqdn.value = "example.org"
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.Registry.Api.Domain.Domain/get_domain_life_cycle_stage", timeout=None)],
        )

    async def test_get_domain_life_cycle_stage_not_exist(self):
        reply = DomainLifeCycleStageReply()
        reply.data.not_registered.CopyFrom(Empty())
        self.client.mock.return_value = make_awaitable(reply)

        with self.assertRaises(DomainDoesNotExist):
            await self.client.get_domain_life_cycle_stage("example.org")

        request = DomainLifeCycleStageRequest()
        request.fqdn.value = "example.org"
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.Registry.Api.Domain.Domain/get_domain_life_cycle_stage", timeout=None)],
        )

    async def test_get_domain_life_cycle_in_auction(self):
        reply = DomainLifeCycleStageReply()
        reply.data.in_auction.CopyFrom(Empty())
        self.client.mock.return_value = make_awaitable(reply)

        with self.assertRaises(DomainDoesNotExist):
            await self.client.get_domain_life_cycle_stage("example.org")

        request = DomainLifeCycleStageRequest()
        request.fqdn.value = "example.org"
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.Registry.Api.Domain.Domain/get_domain_life_cycle_stage", timeout=None)],
        )

    async def _test_get_life_cycle_stage(
        self,
        reply: DomainLifeCycleStageReply,
        result: DomainLifeCycleStageResult,
        *,
        fqdn: str = "example.org",
        request_fqdn: str = "example.org",
    ) -> None:
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(await self.client.get_life_cycle_stage(fqdn), result)

        request = DomainLifeCycleStageRequest()
        request.fqdn.value = request_fqdn
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.Registry.Api.Domain.Domain/get_domain_life_cycle_stage", timeout=None)],
        )

    async def test_get_life_cycle_stage(self):
        reply = DomainLifeCycleStageReply()
        reply.data.domain_id.uuid.value = "42"
        result = DomainLifeCycleStageResult(DomainLifeCycleStage.REGISTERED, DomainId("42"))
        await self._test_get_life_cycle_stage(reply, result)

    async def test_get_life_cycle_stage_idn(self):
        reply = DomainLifeCycleStageReply()
        reply.data.domain_id.uuid.value = "42"
        result = DomainLifeCycleStageResult(DomainLifeCycleStage.REGISTERED, DomainId("42"))
        await self._test_get_life_cycle_stage(
            reply, result, fqdn="háčkyčárky.cz", request_fqdn="xn--hkyrky-ptac70bc.cz"
        )

    async def test_get_life_cycle_stage_delete_candidate(self):
        reply = DomainLifeCycleStageReply()
        reply.data.delete_candidate.CopyFrom(Empty())
        result = DomainLifeCycleStageResult(DomainLifeCycleStage.DELETE_CANDIDATE, None)
        await self._test_get_life_cycle_stage(reply, result)

    async def test_get_life_cycle_stage_not_exist(self):
        reply = DomainLifeCycleStageReply()
        reply.data.not_registered.CopyFrom(Empty())
        result = DomainLifeCycleStageResult(DomainLifeCycleStage.NOT_REGISTERED, None)
        await self._test_get_life_cycle_stage(reply, result)

    async def test_get_life_cycle_in_auction(self):
        reply = DomainLifeCycleStageReply()
        reply.data.in_auction.CopyFrom(Empty())
        result = DomainLifeCycleStageResult(DomainLifeCycleStage.IN_AUCTION, None)
        await self._test_get_life_cycle_stage(reply, result)

    async def test_get_life_cycle_blacklisted(self):
        reply = DomainLifeCycleStageReply()
        reply.data.blacklisted.CopyFrom(Empty())
        result = DomainLifeCycleStageResult(DomainLifeCycleStage.BLACKLISTED, None)
        await self._test_get_life_cycle_stage(reply, result)

    async def test_get_life_cycle_stage_empty(self):
        reply = DomainLifeCycleStageReply()
        result = DomainLifeCycleStageResult(DomainLifeCycleStage.NOT_REGISTERED, None)
        await self._test_get_life_cycle_stage(reply, result)


class DomainAdminClientTest(IsolatedAsyncioTestCase):
    """Test DomainAdminClient's methods."""

    def setUp(self):
        self.client = TestDomainAdminClient(netloc=sentinel.netloc)

    async def _test_batch_delete_domains(
        self,
        domains: Sequence[Domain],
        reply: BatchDeleteDomainsReply,
        results: List[Dict],
        requests: List[BatchDeleteDomainsRequest],
        **kwargs: Any,
    ) -> None:
        self.client.mock.return_value = aiter([reply])

        undeleted = [i async for i in self.client.batch_delete_domains(domains, **kwargs)]

        self.assertEqual(undeleted, results)
        self.client.mock.assert_called_once_with(
            tuple(requests),
            method="/Fred.Registry.Api.Domain.Admin/batch_delete_domains",
            timeout=None,
        )

    async def test_batch_delete_domains_empty_result(self):
        domains = [
            Domain(
                domain_id=DomainId("holly{}".format(i)),
                fqdn="example.org",
                domain_history_id=DomainHistoryId("hilly{}".format(i)),
            )
            for i in range(3)
        ]
        request = BatchDeleteDomainsRequest()
        for domain in domains:
            message = BatchDeleteDomainsRequest.Domain()
            message.domain_id.uuid.value = domain.domain_id
            message.domain_history_id.uuid.value = domain.domain_history_id
            request.domains.append(message)

        await self._test_batch_delete_domains(
            domains=domains,
            reply=BatchDeleteDomainsReply(),
            results=[],
            requests=[request],
        )

    async def test_batch_delete_domains(self):
        reply = BatchDeleteDomainsReply()
        domain = BatchDeleteDomainsReply.NotDeleted()
        domain.domain_id.uuid.value = "holly"
        domain.reason = NotDeletedReason.DOMAIN_BLOCKED.value
        reply.data.not_deleted_domains.append(domain)
        domains = [
            Domain(
                domain_id=DomainId("holly{}".format(i)),
                fqdn="example.org",
                domain_history_id=DomainHistoryId("hilly{}".format(i)),
            )
            for i in range(3)
        ]
        request = BatchDeleteDomainsRequest()
        for domain in domains:
            message = BatchDeleteDomainsRequest.Domain()
            message.domain_id.uuid.value = domain.domain_id
            message.domain_history_id.uuid.value = domain.domain_history_id
            request.domains.append(message)

        await self._test_batch_delete_domains(
            domains=domains,
            reply=reply,
            results=[{"domain_id": "holly", "reason": NotDeletedReason.DOMAIN_BLOCKED}],
            requests=[request],
        )

    async def test_batch_delete_domains_batch_size(self):
        domains = [
            Domain(
                domain_id=DomainId("holly{}".format(i)),
                fqdn="example.org",
                domain_history_id=DomainHistoryId("hilly{}".format(i)),
            )
            for i in range(3)
        ]
        requests = []
        for domain in domains:
            request = BatchDeleteDomainsRequest()
            message = BatchDeleteDomainsRequest.Domain()
            message.domain_id.uuid.value = domain.domain_id
            message.domain_history_id.uuid.value = domain.domain_history_id
            request.domains.append(message)
            requests.append(request)

        await self._test_batch_delete_domains(
            domains=domains,
            reply=BatchDeleteDomainsReply(),
            results=[],
            requests=requests,
            batch_size=1,
        )

    async def test_batch_delete_domains_invalid(self):
        # Test call with invalid Domain object - missing domain_history_id.
        domains = [Domain(domain_id=DomainId("holly"), fqdn="example.org", domain_history_id=None)]
        self.client.mock.return_value = aiter([BatchDeleteDomainsReply()])

        with self.assertRaisesRegex(ValueError, "is missing a domain_history_id"):
            await alist(self.client.batch_delete_domains(domains))

        self.client.mock.assert_not_called()

    async def test_manage_domain_state_flags_empty_domains(self):
        self.client.mock.return_value = aiter([])

        response = self.client.manage_domain_state_flags(domains=[], state_flags={"linked": None})

        self.assertEqual(await alist(response), [])
        self.client.mock.assert_called_once_with(
            (),
            method="/Fred.Registry.Api.Domain.Admin/manage_domain_state_flags",
            timeout=None,
        )

    async def test_manage_domain_state_flags_empty_state_flags(self):
        self.client.mock.return_value = aiter([])
        domains = [
            Domain(domain_id=DomainId("holly:1"), fqdn="holly.example.org"),
        ]

        response = self.client.manage_domain_state_flags(domains=domains, state_flags={})

        self.assertEqual(await alist(response), [])

        request = ManageDomainStateFlagsRequest()
        for domain in domains:
            dom = request.state_flag_requests_to_create.domains.add()
            dom.domain_id.uuid.value = domain.domain_id
        self.client.mock.assert_called_once_with(
            (request,),
            method="/Fred.Registry.Api.Domain.Admin/manage_domain_state_flags",
            timeout=None,
        )

    async def test_manage_domain_state_flags_close(self):
        request_ids = ["request:1", "request:2"]
        self.client.mock.return_value = aiter([ManageDomainStateFlagsReply()])

        response = self.client.manage_domain_state_flags(close_requests=request_ids)

        self.assertEqual(await alist(response), [])
        request = ManageDomainStateFlagsRequest()
        for rid in request_ids:
            flag = request.state_flag_requests_to_close.add()
            flag.uuid.value = rid
        self.client.mock.assert_called_once_with(
            (request,),
            method="/Fred.Registry.Api.Domain.Admin/manage_domain_state_flags",
            timeout=None,
        )

    async def test_manage_domain_state_flags_set(self):
        valid_from = datetime(2022, 1, 1, tzinfo=pytz.utc)
        valid_to = datetime(2022, 2, 1, tzinfo=pytz.utc)
        domains = [
            Domain(domain_id=DomainId("holly:1"), fqdn="holly.example.org"),
            Domain(
                domain_id=DomainId("hilly:1"), fqdn="hilly.example.org", domain_history_id=DomainHistoryId("hilly:2")
            ),
        ]
        state_flags = {
            "linked": None,
            "serverBlocked": DatetimeRange(start=valid_from, end=valid_to),
        }
        reply = ManageDomainStateFlagsReply()
        for i, domain in enumerate(domains):
            for j, flag in enumerate(state_flags.keys()):
                info = reply.data.state_flag_requests_created.add()
                info.request_id.uuid.value = f"request:{i*10+j}"
                info.domain_id.uuid.value = domain.domain_id
                info.state_flag = flag
        self.client.mock.return_value = aiter([reply])

        response = await alist(self.client.manage_domain_state_flags(domains=domains, state_flags=state_flags))

        self.assertEqual(
            response,
            [
                {"domain_id": DomainId("holly:1"), "request_id": "request:0", "state_flag": "linked"},
                {"domain_id": DomainId("holly:1"), "request_id": "request:1", "state_flag": "serverBlocked"},
                {"domain_id": DomainId("hilly:1"), "request_id": "request:10", "state_flag": "linked"},
                {"domain_id": DomainId("hilly:1"), "request_id": "request:11", "state_flag": "serverBlocked"},
            ],
        )
        self.assertIsInstance(response[0]["domain_id"], DomainId)
        self.assertIsInstance(response[1]["domain_id"], DomainId)

        request = ManageDomainStateFlagsRequest()
        for domain in domains:
            dom = request.state_flag_requests_to_create.domains.add()
            dom.domain_id.uuid.value = domain.domain_id
            if domain.domain_history_id:
                dom.domain_history_id.uuid.value = domain.domain_history_id
        request.state_flag_requests_to_create.state_flags.get_or_create("linked")
        validity = request.state_flag_requests_to_create.state_flags.get_or_create("serverBlocked")
        validity.valid_from.FromDatetime(valid_from)
        validity.valid_to.FromDatetime(valid_to)

        self.client.mock.assert_called_once_with(
            (request,),
            method="/Fred.Registry.Api.Domain.Admin/manage_domain_state_flags",
            timeout=None,
        )

    async def test_manage_domain_state_flags_mismatch(self):
        domains = [Domain(domain_id=DomainId("holly"), fqdn="example.org")]
        state_flags = {"linked": None}

        with self.assertRaises(ValueError):
            await alist(self.client.manage_domain_state_flags(domains=domains))  # type: ignore[call-overload]
        with self.assertRaises(ValueError):
            await alist(self.client.manage_domain_state_flags(state_flags=state_flags))  # type: ignore[call-overload]

    async def _test_get_domains_notify_info(self, method: str, start: datetime, end: datetime, page_size: int) -> None:
        reply = GetDomainsNotifyInfoReply()
        for i in range(page_size):
            domain = reply.data.domains.add()
            domain.domain_id.uuid.value = f"domain.{i}"
            domain.fqdn = f"domain-{i}.cz"
            domain.registrant_id.uuid.value = f"contact.{i}"
            domain.registrant_handle.value = f"CONTACT.{i}"
            domain.registrant_name = f"Mr. {i}"
            domain.registrant_organization = f"{i} corp."
            domain.contacts.add(email=f"contact@domain-{i}.cz")
            domain.additional_contacts.add(email=f"additional@domain-{i}.cz")
            domain.additional_contacts.add(telephone=f"+420.{i}")
        reply.data.pagination.next_page_token = "NEXT"
        reply.data.pagination.items_left = 7

        self.client.mock.return_value = aiter([reply])
        response = await getattr(self.client, method)(
            start=start,
            end=end,
            page_size=page_size,
        )
        self.assertEqual(
            response,
            {
                "domains": [
                    {
                        "domain_id": f"domain.{i}",
                        "fqdn": f"domain-{i}.cz",
                        "registrant_id": f"contact.{i}",
                        "registrant_handle": f"CONTACT.{i}",
                        "registrant_name": f"Mr. {i}",
                        "registrant_organization": f"{i} corp.",
                        "contacts": [ContactInfo(type=ContactInfoType.EMAIL, value=f"contact@domain-{i}.cz")],
                        "additional_contacts": [
                            ContactInfo(type=ContactInfoType.EMAIL, value=f"additional@domain-{i}.cz"),
                            ContactInfo(type=ContactInfoType.TELEPHONE, value=f"+420.{i}"),
                        ],
                    }
                    for i in range(page_size)
                ],
                "pagination": {"next_page_token": "NEXT", "items_left": 7},
            },
        )

    async def test_get_domains_outzone_notify_info(self):
        start = datetime(2022, 1, 1, tzinfo=pytz.utc)
        end = datetime(2022, 1, 2, tzinfo=pytz.utc)

        await self._test_get_domains_notify_info("get_domains_outzone_notify_info", start, end, 5)

        request = GetDomainsOutzoneNotifyInfoRequest()
        request.outzone_from.FromDatetime(start)
        request.outzone_to.FromDatetime(end)
        request.pagination.page_size = 5

        self.assertEqual(
            self.client.mock.mock_calls,
            [
                call(request, method="/Fred.Registry.Api.Domain.Admin/get_domains_outzone_notify_info", timeout=None),
            ],
        )

    async def test_get_domains_delete_notify_info(self):
        start = datetime(2022, 1, 1, tzinfo=pytz.utc)
        end = datetime(2022, 1, 2, tzinfo=pytz.utc)

        await self._test_get_domains_notify_info("get_domains_delete_notify_info", start, end, 10)

        request = GetDomainsDeleteNotifyInfoRequest()
        request.delete_from.FromDatetime(start)
        request.delete_to.FromDatetime(end)
        request.pagination.page_size = 10

        self.assertEqual(
            self.client.mock.mock_calls,
            [
                call(request, method="/Fred.Registry.Api.Domain.Admin/get_domains_delete_notify_info", timeout=None),
            ],
        )

    async def test_update_domains_outzone_additional_notify_info(self):
        self.client.mock.return_value = make_awaitable(UpdateDomainsAdditionalNotifyInfoReply())

        start = datetime(2022, 1, 1, tzinfo=pytz.utc)
        end = datetime(2022, 1, 2, tzinfo=pytz.utc)
        domains = {
            DomainId(f"domain.{i}"): [
                ContactInfo(type=ContactInfoType.EMAIL, value="email@example.com"),
                ContactInfo(type=ContactInfoType.TELEPHONE, value="+420.000000"),
            ]
            for i in range(3)
        }

        await self.client.update_domains_outzone_additional_notify_info(
            domains,
            start=start,
            end=end,
            batch_size=2,
        )

        requests = []
        domain_items = tuple(domains.items())
        for request_domains in domain_items[:2], domain_items[2:]:
            request = UpdateDomainsOutzoneAdditionalNotifyInfoRequest()
            for domain_id, contacts in request_domains:
                domain = request.domains_additional_info.add()
                domain.domain_id.uuid.value = domain_id
                domain.outzone_from.FromDatetime(start)
                domain.outzone_to.FromDatetime(end)
                domain.additional_contacts.add(email="email@example.com")
                domain.additional_contacts.add(telephone="+420.000000")
            requests.append(request)
        self.assertEqual(
            self.client.mock.mock_calls,
            [
                call(
                    tuple(requests),
                    method="/Fred.Registry.Api.Domain.Admin/update_domains_outzone_additional_notify_info",
                    timeout=None,
                ),
            ],
        )

    async def test_update_domains_delete_additional_notify_info(self):
        self.client.mock.return_value = make_awaitable(UpdateDomainsAdditionalNotifyInfoReply())

        start = datetime(2022, 1, 1, tzinfo=pytz.utc)
        end = datetime(2022, 1, 2, tzinfo=pytz.utc)
        domains = {
            DomainId(f"domain.{i}"): [
                ContactInfo(type=ContactInfoType.EMAIL, value="email@example.com"),
                ContactInfo(type=ContactInfoType.TELEPHONE, value="+420.000000"),
            ]
            for i in range(3)
        }

        await self.client.update_domains_delete_additional_notify_info(
            domains,
            start=start,
            end=end,
            batch_size=2,
        )

        requests = []
        domain_items = tuple(domains.items())
        for request_domains in domain_items[:2], domain_items[2:]:
            request = UpdateDomainsDeleteAdditionalNotifyInfoRequest()
            for domain_id, contacts in request_domains:
                domain = request.domains_additional_info.add()
                domain.domain_id.uuid.value = domain_id
                domain.delete_from.FromDatetime(start)
                domain.delete_to.FromDatetime(end)
                domain.additional_contacts.add(email="email@example.com")
                domain.additional_contacts.add(telephone="+420.000000")
            requests.append(request)
        self.assertEqual(
            self.client.mock.mock_calls,
            [
                call(
                    tuple(requests),
                    method="/Fred.Registry.Api.Domain.Admin/update_domains_delete_additional_notify_info",
                    timeout=None,
                ),
            ],
        )
